k : name_type.

pred p k.

(* Unary properties *)
?(forall a. p a) => (forall a. p a). (* valid *)
?(forall a. p a) => (exists a. p a). (* valid *)
?(forall a. p a) => (new a. p a). (* valid *)
?(new a. p a) => (forall a. p a). (* classicallly valid *)
?(new a. p a) => (exists a. p a). (* valid X *)
?(new a. p a) => (new a. p a).    (* valid *)


(* Binary properties *)
pred q k k.

?(forall a. q a a) => (forall a. q a a). (* valid *)

?(forall a,b. q a b) => (forall a,b. q b a). (* valid *)

?(forall a. q a a) => (forall a,b. q a b). (* invalid *)

?(forall a,b. q a b) => (forall a. q a a). (* valid *)

?(forall a. q a a) => (exists a. q a a). (* valid *)

?(forall a,b. q a b) => (exists a,b. q a b). (* valid *)

?(forall a. q a a) => (exists a,b. q a b). (* valid *)

?(forall a,b. q a b) => (exists a. q a a). (* valid *)

?(forall a. q a a) => (new a. q a a). (* valid *)

?(forall a,b. q a b) => (new a,b. q a b). (* valid *)

?(forall a. q a a) => (new a,b. q a b). (* invalid *)

?(forall a,b. q a b) => (new a. q a a). (* valid *)


?(new a. q a a) => (forall a. q a a). (* classically valid X*)

?(new a,b. q a b) => (forall a,b. q b a). (* invalid *)

?(new a. q a a) => (forall a,b. q a b). (* invalid *)

?(new a,b. q a b) => (forall a. q a a). (* invalid *)

?(new a. q a a) => (exists a. q a a). (* valid X*)

?(new a,b. q a b) => (exists a,b. q a b). (* valid X*)

?(new a. q a a) => (exists a,b. q a b). (* valid X*)

?(new a,b. q a b) => (exists a. q a a). (* invalid *)

?(new a. q a a) => (new b. q b b). (* valid *)

?(new a, b. q a b) => (new a,b. q b a). (* valid *)

?(new a. q a a) => (new a,b. q b a).    (* invalid *)

?(new a,b. q b a) => (new a. q a a).   (* invalid *)
 



(* Freshness properties *)

?forall A. new b. b # A. (* valid (= freshness axiom ) *)

?exists A. new b. b # A. (* valid iff A nonempty *)

?new a. forall B. a # B. (* invalid: B = a !! 
			    succeeds anyway because we need to check that
			    constraints mentioning B don't escape B's scope. 
			    how to check this: at the end of a new a, filter 
			    out constraints a # X where X was in scope 
			    when the new occurred.  similarly, at the end of
			    an exists B, filter out any unsolved a # B's where 
			    a already is in scope.  complain if any freshness 
			    constraints mentioning A are extant at the end of 
                            a forall A.  *)

?new a. exists B. a # B. (* valid iff B nonempty *)
 
?new a. new b.    a # b. (* valid *)


(* Equality properties *)
?forall A. forall B. B = A. (* invalid *)

?forall A. exists B. B = A. (* valid *)

?forall A. new b. b = A. (* invalid *)

?exists A. forall B. B = A. (* invalid: *)

?exists A. exists B. B = A. (* valid *)

?exists A. new b. b = A. (* invalid: 
			    succeeds because we need to check that b # A when 
			    discharging "new". how to do this?
			    one way: same mechanism as for uvars. 
			    remember atoms that are allowed to occur in A,
			    and complain if some appear in A 
			    that aren't in the list *)

?new a. forall B. a = B. (* invalid: B = a *)

?new a. exists B. a = B. (* valid *)

?new a. new b.    a = b. (* invalid *)


(* Permutations & freshness *)
?exists Y. X = (a~b)Y, Y = a, X = (a~c)Y.
?X = (a~b)X, X = a.
?X = (a~b)X, X # a.
?a\X = b\Y, (a~b)X = (a~b)Y.

(* Numerics *)

func fibonacci int = int.

fibonacci(N) = F :- (
  N < 2  -> 
    F = 1 
  | F is fibonacci(N - 1) + fibonacci(N - 2)
).

? X = fibonacci 10.

? X is fibonacci 10.

pred factorial int int.

factorial 0 0.
factorial 1 1.
factorial N M :- 
	2 <= N,
	A is N - 1,
	factorial A B,
	M is N * B.

? factorial 10 X.

