open Types;;
module P = Perm;;
open Internal;;
module S = Subst;;
open Printer;;

type 'a fconstrs = ('a term * var) list;;
type 'a evconstrs = (('a term * 'a term) list) list;;
type 'a answer = ('a S.subst * 'a fconstrs * 'a evconstrs);;
type 'a sc = 'a answer -> unit

exception EUnify;;



let compress subst t = 
  match t with
    Susp(pi,_,x) -> 
      (match S.lookup subst x with
	Some t -> apply_p pi t
      |	None -> t)
  | _ -> t
;;
	
let pp_evconstrs pp evcs = 
  let pp_evconstr evc =
    let lc,rc = Util.unzip evc in 
    paren (sep comma (List.map (pp_term pp) lc)) 
      <+> tilde <:> eq <+> 
    paren (sep comma (List.map (pp_term pp) rc))
  in 
  vcat (List.map pp_evconstr evcs)


(* step 1. Remove all function applications and abstractions
   resulting in a list of suspensions and atoms.
   step 2. Build an equivalence relation consisting of the reflexive transitive
   closure of pointwise equality on both sides.  
   step 3. Unify all the equivalence classes; fail if any classes fail to 
   unify, otherwise build a new tuple out of the classes and add it as 
   a constraint. *)

(* TODO: This is broken because substitutions that occur during 
unification won't reactivate delayed ev constraints. Serious problem! 
Need to reorganize code.  *)

let new_var() = Susp(Perm.id,Exist Toplevel,Var.mkvar "_");;

let eunify aeq t u sc ans = 
  let rec classify t u sc classes ans = 
    let (subst,fconstrs,evconstrs) = ans in 
    match classes with
      [] -> sc [(t,u)] ans
    | (t',u')::classes -> 
	if (Internal.eq aeq t t') || (Internal.eq aeq u u') 
	then Unify.unifys aeq [(t,t');(u,u')] (fun (subst,fconstrs) -> 
	  classify t u sc classes (subst,fconstrs,evconstrs)) (subst,fconstrs)
	else classify t u (fun classes ans -> 
	  sc ((t',u')::classes) ans) 
	    classes ans
  in 
  let rec eunify' t u sc classes ans = 
  let (subst,fconstrs,evconstrs) = ans in 
    match compress subst t, compress subst u with
    | App(f,ts'),u -> 
	let vs = List.map (fun _ -> new_var()) ts' in 
	Unify.unify aeq (App(f,vs)) u 
	  (fun (subst,constrs) -> 
	    eunifys' ts' vs sc classes (subst,constrs,evconstrs)) 
	  (subst,fconstrs)
    | t, App(f,us') -> 
	let vs = List.map (fun _ -> new_var()) us' in 
	Unify.unify aeq (App(f,vs)) t 
	  (fun (subst,fconstrs) -> 
	    eunifys' vs us' sc classes (subst,fconstrs,evconstrs)) 
	  (subst,fconstrs)
    |  Abs(a,t),u -> 
	let c = Var.rename a in
	let x = new_var() in 
	Unify.unify aeq (Abs(c,x)) u (fun (subst,fconstrs) -> 
	  eunifys' [Name(a);t] [Name(c);x] sc classes (subst,fconstrs,evconstrs))
	  (subst,fconstrs)
    |  t,Abs(a,u) -> 
	let c = Var.rename a in
	let x = new_var() in 
	Unify.unify aeq (Abs(c,x)) t (fun (subst,fconstrs) -> 
	  eunifys' [Name(c);x] [Name(a);u] sc classes (subst,fconstrs,evconstrs))
	  (subst,fconstrs)
    | t',u' -> classify t' u' sc classes ans
  and eunifys' ts us sc classes ans = 
    match ts,us with
      [],[] -> sc classes ans
    | t::ts',u::us' -> eunify' t u (eunifys' ts' us' sc) classes ans
    | _,_ -> ()
  in 
  (* TODO: Turn classes into constraint *)
  eunify' t u (fun classes (subst,fconstrs,evconstrs) -> 
    sc (subst,fconstrs,classes::evconstrs)) [] ans
